/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package campus;


import java.io.FileReader;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.Iterator;
import java.util.Hashtable;
import java.util.StringTokenizer;


/**
 *  Contains mapping information about APs and buidings.
 *  
 *  @author     Guanling Chen
 *  @author     Yurong Xu
 *  @version    $Id: CampusMap.java,v 1.4 2003/09/13 03:52:41 glchen Exp $
 */
public class CampusMap 
{
    public static final String APCONFIG     = "apconfig.map";
    public static final String APLOCATE     = "aplocate.map";
    public static final String BUILDINGS    = "buildings.map";

    protected String _mappath;
    protected Hashtable _accesspoints = new Hashtable();
    protected Hashtable _buildings = new Hashtable();

    
    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public CampusMap(String path)
        throws Exception
    {
        _mappath = path;
        if (!_mappath.endsWith("/"))
        {
            _mappath = _mappath + "/";
        }
        loadBuilding();
        loadAccessPoints();
    }

    private void loadBuilding()
        throws Exception
    {
        BufferedReader in = new BufferedReader(new FileReader(_mappath + BUILDINGS));
        String line = null;
        while((line = in.readLine()) != null)
        {
            Building building = new Building();
            StringTokenizer st = new StringTokenizer(line, ",");
            if (st.hasMoreTokens()) { building.setCode(convert(st.nextToken())); }
            if (st.hasMoreTokens()) { building.setCategory(convert(st.nextToken())); }
            if (st.hasMoreTokens()) { building.setName(convert(st.nextToken())); }
            if (st.hasMoreTokens()) { building.setStreet(convert(st.nextToken())); }
            if (st.hasMoreTokens()) { building.setCity(convert(st.nextToken())); }
            if (st.hasMoreTokens()) { building.setState(convert(st.nextToken())); }
            if (st.hasMoreTokens()) { building.setZipCode(convert(st.nextToken())); }
            if (st.hasMoreTokens()) { building.setUrl(convert(st.nextToken())); }
            building.setCountry("US");
            if (building.getStreet() != null 
                || building.getCity() != null)
            {
                building.setCertainty(95);
            }
            else
            {   // get default if not present in map file
                building.setCity("Hanover");
                building.setState("NH");
                building.setZipCode("03755");
                building.setCertainty(90);
                building.setCityUrl("http://www.hanovernh.org/");
            }
            _buildings.put(building.getCode(), building);
        }
        in.close();
    }

    private void loadAccessPoints()
        throws Exception
    {
        // read APLOCATE data file first
        BufferedReader in = new BufferedReader(
            new FileReader(_mappath + APLOCATE));
        String line = null;
        while((line = in.readLine()) != null)
        {
            String[] data = line.split(",");
            String apname = data[0];
            AccessPoint ap = (AccessPoint) _accesspoints.get(apname);
            if (ap == null) 
            {
                ap = new AccessPoint();
                ap.setName(apname);
            }
            ap.setX(Double.parseDouble(data[1]));
            ap.setY(Double.parseDouble(data[2]));
            ap.setZ(Double.parseDouble(data[3]));
            ap.setFloor(Integer.parseInt(data[4].substring(2)));
            _accesspoints.put(apname, ap);
        }
        in.close();
        // now read the APCONFIG data file
        in = new BufferedReader(new FileReader(_mappath + APCONFIG));
        while((line = in.readLine()) != null)
        {
            int p1 = 0;
            int p2 = line.indexOf(",", p1);
            String building = line.substring(p1, p2);
            p1 = p2 + 1;
            p2 = line.indexOf(",", p1);
            String antenna = line.substring(p1, p2);
            p1 = p2 + 1;
            p2 = line.indexOf(",", p1);
            String model = line.substring(p1, p2);
            p1 = p2 + 1;
            p2 = line.indexOf(",", p1);
            String info = line.substring(p1, p2);
            p1 = p2 + 1;
            p2 = line.indexOf(",", p1);
            String channel = line.substring(p1, p2);
            p1 = p2 + 1;
            p2 = line.indexOf(",", p1);
            String power = line.substring(p1, p2);
            p1 = p2 + 1;
            p2 = line.indexOf(",", p1);
            String apname = line.substring(p1, p2);
            if (apname.equals(""))
            {
//                System.err.println("CampusMap: ignore line [" + line + "]");
                continue;
            }
            p1 = p2 + 1;
            p2 = line.indexOf(",", p1);
            String ipaddress = line.substring(p1, p2);
            p1 = p2 + 1;
            p2 = line.indexOf(",", p1);
            String gateway = line.substring(p1, p2);
            p1 = p2 + 1;
            String macaddress = line.substring(p1);
            AccessPoint ap = (AccessPoint) _accesspoints.get(apname);
            if (ap == null) 
            {
                ap = new AccessPoint();
                ap.setName(apname);
            }
            ap.setBuilding(building);
            ap.setAntenna(antenna);
            ap.setModel(model);
            ap.setInfo(info);
            ap.setChannel(channel);
            ap.setPower(power);
            ap.setIpAddress(ipaddress);
            ap.setMacAddress(macaddress);
            _accesspoints.put(apname, ap);
//            System.out.println(building + "," + antenna + "," + model + "," 
//                + place + "," + channel + "," + power + "," + apname + ","
//                + ipaddress + "," + gateway + "," + macaddress);
        }
        in.close();
    }

    private String convert(String token)
    {
        if (token.equals("?"))
        {
            return null;
        }
        else
        {
            return token;
        }
    }
    
    //=======================================================================
    // Public interface methods.
    //=======================================================================

    public Building getBuildingByAp(String apname)
    {
        return (Building) _buildings.get(AccessPoint.getBuildingCode(apname));
    }

    public void printAccessPoints()
    {
        Iterator itr = _accesspoints.values().iterator();
        while (itr.hasNext())
        {
            AccessPoint ap = (AccessPoint) itr.next();
            String code = AccessPoint.getBuildingCode(ap.getName());
            System.out.println(code + "," + ap.getBuilding());
        }
    }


    //=======================================================================
    // Test
    //=======================================================================

    public static void main(String[] args)
        throws Exception
    {
        String apname = (args.length > 0) ? args[0] : "baker11-ap";
        CampusMap map = new CampusMap("/u/glchen/work/solar/src/java/campus/");
//        map.printAccessPoints();
        BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
        String line = null;
        while ((line = in.readLine()) != null)
        {
            System.out.println(AccessPoint.getBuildingCode(line));
        }
        in.close();

    }
}
