/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package campus.compass;


import org.apache.log4j.Logger;
import org.doit.muffin.CompassMain;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.Socket;
import java.net.InetAddress;
import java.util.Iterator;
import java.util.Hashtable;

import solar.service.BaseService;
import solar.service.ServiceException;
import solar.util.NetUtil;
import campus.Location;
import campus.CampusMap;
import campus.compass.WhoisSource;
import campus.compass.GeobytesSource;
import campus.compass.ZipCodeSource;
import campus.snmp.ApStat;
import campus.snmp.SnmpLocator;
import campus.syslog.SyslogParser;


/**
 *  Compass service provides a Web proxy that impose location-dependent 
 *  information on the pages viewed by the clients.  The implementation
 *  is based on Muffin (http://muffin.doit.org/).
 *  
 *  @author     Guanling Chen
 *  @version    $Id: CompassService.java,v 1.8 2003/09/21 03:13:35 glchen Exp $
 */
public class CompassService 
    extends BaseService
    implements Runnable
{
    public static final String VERSION = "0.1";
    protected static Logger _log = Logger.getLogger(CompassService.class);

    protected static String _locfile;
    protected static Hashtable _loctable = new Hashtable();

    protected String _sysloghost;
    protected int _syslogport;
    protected String _snmpscript;
    protected String _community;
    protected String _mappath;
    protected static SnmpLocator _snmploc;
    protected static CampusMap _map;
    protected static boolean _needparse;
    protected static SyslogParser _parser = new SyslogParser(2003);
    
    
    //=======================================================================
    //  Implementation of solar.service.Service interface.
    //=======================================================================
    
    public void init()
        throws ServiceException
    {
        // options are: nogui, 
        serp.util.Options opts = 
            getServiceManager().getServiceConfig(getName());

        // start snmp locator if specified, only works for wireless IP
        try
        {
            _needparse = opts.getBooleanProperty("syslogparser", false);
            _mappath = opts.getProperty("mappath");
            _community = opts.getProperty("community");
            _sysloghost = opts.getProperty("sysloghost");
            _syslogport = opts.getIntProperty("syslogport");
            _snmpscript = opts.getProperty("snmpscript");
            if (_community != null && _mappath != null 
                && _sysloghost != null && _snmpscript != null)
            {
                _map = new CampusMap(_mappath);
                _snmploc = new SnmpLocator(_snmpscript, _community);
                _log.info("SNMP locator started.");
                new Thread(this).start();   // read the fifo
            }
        }
        catch (Exception ex)
        {
            throw new ServiceException(ex);
        }

        // start a general IP locator then
        _locfile = opts.getProperty("iplocator");
        if (_locfile == null)
        {
            throw new ServiceException("Fail to init: iplocator not specified.");
        }
        try
        {
            BufferedReader in = new BufferedReader(new FileReader(_locfile));
            String line = in.readLine();
            while (line != null)
            {
                String[] info = line.split(",");
                String ip = info[0];
                Location loc = new Location();
                if (info[1].length() > 0) loc.setStreet(info[1]);
                if (info[2].length() > 0) loc.setCity(info[2]);
                if (info[3].length() > 0) loc.setState(info[3]);
                if (info[4].length() > 0) loc.setCountry(info[4]);
                if (info[5].length() > 0) loc.setZipCode(info[5]);
                if (info[6].length() > 0) loc.setCertainty(Integer.parseInt(info[6]));
                _loctable.put(ip, loc);
                line = in.readLine();
            }
            in.close();
        }
        catch (Exception ex)
        {
            throw new ServiceException(ex);
        }
        CompassMain compass = new CompassMain(opts);
    }

    //=======================================================================
    //  Implementation of Runnable interface
    //=======================================================================
    
    public void run()
    {
        try
        {
            Socket socket = new Socket(_sysloghost, _syslogport);
            BufferedReader in = new BufferedReader(
                new InputStreamReader(socket.getInputStream()), 1);
            String line = in.readLine();
            while (line != null)
            {
                try
                {
                    if (_needparse)
                    {
                        _snmploc.update(_parser.parse(line));
                    }
                    else
                    {
                        _snmploc.update(line);
                    }
                }
                catch (Exception ex)
                {
                    ex.printStackTrace();
                    System.err.println(line);
                }
                line = in.readLine();
            }
            in.close();
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
        _log.error("SNMP locator input thread stopped.");
    }

    //=======================================================================
    //  Public interface methods.
    //=======================================================================

    public static ApStat getApStatByIp(String ip)
    {
        if (_snmploc != null)
        {
            String ap = _snmploc.getApByIp(ip);
            if (ap != null)
            {
                return _snmploc.getApStat(ap);
            }
        }
        return null;        
    }
    
    public static String getApByIp(String ip)
    {
        if (_snmploc != null)
        {
            return _snmploc.getApByIp(ip);
        }
        return null;
    }

    public static Location getLocationByIp(String ip)
        throws Exception
    {
        if (ip.equals("127.0.0.1"))
        {
            ip = NetUtil.getIpAddress(InetAddress.getLocalHost());
        }
        if (ip.startsWith("10.") || ip.startsWith("192.169."))
        {
            throw new IllegalArgumentException("Reserved non-routable IP.");
        }
        // lookup the snmp locator first, TODO: should check IP is wireless
        Location ret = null;
        if (_snmploc != null)
        {
            String apname = _snmploc.getApByIp(ip);
            if (apname != null)
            {
                System.err.println("SNMP: " + ip + " at " + apname);
                return _map.getBuildingByAp(apname);
            }
        }
        // otherwise check the cached table laoded from disk
        ret = (Location) _loctable.get(ip);
        if (ret != null)
        {
            return ret;
        }
        // check whether the IP segment belongs to dartmouth
        if (ip.startsWith("129.170."))
        {
            ret = new Location(null,"Hanover","NH","US","03755",85);
            addIpLocation(ip, ret);
            return ret;
        }
        // now have to query several locator
        try
        {
            ret = GeobytesSource.getLocationByIp(ip);
            ret.setZipCode(
                ZipCodeSource.getZipCode(ret.getCity(), ret.getState()));
        }
        catch (Exception ex) {}
        if (ret == null)
        {
            // last resort to use whoise
            ret = WhoisSource.getLocationByIp(ip);
        }
        // update the cache on disk
        addIpLocation(ip, ret);
        return ret;
    }

    private static void addIpLocation(String ip, Location loc)
    {
        _loctable.put(ip, loc);
        try
        {
            PrintWriter pw = new PrintWriter(new FileWriter(_locfile));
            Iterator itr = _loctable.keySet().iterator();
            while (itr.hasNext())
            {
                String key = (String) itr.next();
                Location kloc = (Location) _loctable.get(key);
                pw.println(key + "," + kloc.toLogString());
            }
            pw.flush();
            pw.close();
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }
}
