/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package campus.snmp;


import EDU.oswego.cs.dl.util.concurrent.*;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.net.InetAddress;
import java.util.Set;
import java.util.Arrays;
import java.util.Random;
import java.util.Calendar;
import java.util.HashSet;
import java.util.Hashtable;

import solar.api.Event;
import solar.util.NetUtil;
import solar.util.SystemUtil;
import campus.syslog.SyslogConstants;


/**
 *  This locator monitors the Syslog messages and polls APs to fill the
 *  ip-ap and mac-ap mapping tables.  It polls the AP when a new client
 *  is associate with an AP which has not been polled for a fixed interval.
 *  Otherwise, that AP is put into a waiting list which is processed by
 *  a thread periodically.  That thread also randomly picks one AP from all
 *  seen APs for processing.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: SnmpLocator.java,v 1.7 2003/09/21 03:13:36 glchen Exp $
 */
public class SnmpLocator 
    implements SyslogConstants, Runnable
{
    /** A single snmp polling task. */
    class SnmpTask
        implements Runnable
    {
        SnmpLocator _locator;
        String _apname;
        String _clientmac;
        long _timestamp;    // time in associated msg
        
        public SnmpTask(SnmpLocator locator, 
            String apname, String clientmac, long timestamp)
        {
            _locator = locator;
            _apname = apname;
            _clientmac = clientmac;
            _timestamp = timestamp;
        }

        public void run()
        {
            try
            {
                _locator.doPoll(_apname, _clientmac, _timestamp);
            }
            catch (Exception ex)
            {
                ex.printStackTrace();
            }
        }
    }
    
    public static final long POLL_INTERVAL = 4*60*1000;   // 4 minutes
    public static final long WAIT_INTERVAL = 1*60*1000;   // 1 minute

    protected Random _random;
    protected Hashtable _apstat = new Hashtable();
    protected Hashtable _ipLocator = new Hashtable(); // (ip, apname)
    protected Hashtable _macLocator = new Hashtable(); // (mac, apname)

    protected PooledExecutor _pool;
    protected Set _aplist = new HashSet(); // (apname)
    protected Set _working = new HashSet(); // (apname)
    protected Set _waiting = new HashSet(); // (apname)
    protected Hashtable _polled = new Hashtable(); // (apname, last_poll_time)
    protected Object _lock = new Object();
    protected String _community;
    protected String _script;
    
    
    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public SnmpLocator(String community)
    {
        this("/u/glchen/work/script/snmp/poll_one.pl", community);
    }

    public SnmpLocator(String script, String community)
    {
        _random = new Random(System.currentTimeMillis());
        _script = script;
        _community = community;
        _pool = new PooledExecutor(new LinkedQueue(), 9);
        _pool.setKeepAliveTime(-1); // live forever
        _pool.createThreads(9);
        new Thread(this).start();
    }


    //=======================================================================
    // Public static interface.
    //=======================================================================

    /** Update the locator with a parsed syslog message. */
    public void update(String msg)
    {
        String[] data = msg.split("\t");
        long timestamp = Long.parseLong(data[IDX_TIMESTAMP]) * 1000;
        String msgtype = data[IDX_MSGTYPE];
        String apname = data[IDX_APNAME];
        String clienthost = data[IDX_CLIENTHOST];
        String clientmac = data[IDX_CLIENTMAC];
        String action = data[IDX_ACTION];
        doUpdate(timestamp, msgtype, apname, clienthost, clientmac, action);
    }

    /** Update the locator with a published syslog event. */
    public void update(Event evt)
    {
        long timestamp = Long.parseLong((String)evt.getValue(EVT_TIMESTAMP)) * 1000;
        String msgtype = (String)evt.getValue(EVT_MSGTYPE);
        String apname = (String)evt.getValue(EVT_APNAME);
        String clienthost = (String)evt.getValue(EVT_CLIENTHOST);
        String clientmac = (String)evt.getValue(EVT_CLIENTMAC);
        String action = (String)evt.getValue(EVT_ACTION);
        doUpdate(timestamp, msgtype, apname, clienthost, clientmac, action);
    }

    /** Return the current AP of the specified MAC. */
    public synchronized String getApByMac(String mac)
    {
        return (String) _macLocator.get(mac);
    }

    /** Return the current AP of the specified IP address. */
    public synchronized String getApByIp(String ipaddress)
    {
        return (String) _ipLocator.get(ipaddress);
    }

    /** Return some statistics of the specified AP. */
    public ApStat getApStat(String apname)
    {
        synchronized (_lock)
        {
            return (ApStat) _apstat.get(apname);
        }
    }

    
    //=======================================================================
    // Private helper methods.
    //=======================================================================

    /** update the locator with a syslog message. */
    private void doUpdate(long timestamp,
        String msgtype,
        String apname,
        String clienthost,
        String clientmac,
        String action)
    {
        // ignore non-info and non-associated messages
        if ((!msgtype.equals(TAG_INFO)) 
            || (!action.equals(TAG_ASSOCIATED) 
                && !action.equals(TAG_REASSOCIATED)))
        {
            return;
        }
        // ignore the clients who are actually APs
        if (clienthost.endsWith("-ap") || clienthost.endsWith("-wb"))
        {
            return;
        }
        // clean the clienthost, such as removing ?
        //clienthost = clienthost.replaceAll("#", "\\032");
        if (clienthost.indexOf("#") != -1)
        {
            return;
        }
        if (clienthost.startsWith("?"))
        {
            clienthost = clienthost.substring(1);
        }
        /*
        // if an ip address exists, then we are done
        if (clienthost.startsWith("129.170."))
        {
            record(clientmac, clienthost, apname);
            return;
        }
        // if a hostname exists, look it up and we are done
        if (clienthost.endsWith(".Kiewit"))
        {
            try
            {
                String ipaddress = NetUtil.getIpAddress(
                    InetAddress.getByName(clienthost + ".Dartmouth.EDU"));
                record(clientmac, ipaddress, apname);
                return;
            }
            catch (Exception ex)
            {
                // do nothing, fall through to poll AP
            }
        }*/
        // if we seen this mac on this AP last time, do nothing about it
        if (apname.equals(_macLocator.get(clientmac)))
        {
            return;
        }
        // ok to poll now?
        if (toPoll(apname, clientmac))
        {
            addPollTask(apname, clientmac, timestamp);
        }
    }

    /** Check whether an AP is pollable at this time. */
    private boolean toPoll(String apname, String clientmac)
    {
        synchronized (_lock)
        {
            // if the ap is already in the working set, do nothing
            if (_working.contains(apname))
            {
                return false;
            }
            // check if we can poll the AP now, if not, put it to waiting list
            Long pt = (Long) _polled.get(apname);
            if (pt != null)
            {
                if (System.currentTimeMillis() - pt.longValue() < POLL_INTERVAL)
                {
                    if (!_waiting.contains(apname))
                    {
                        _waiting.add(apname);
                        if (clientmac != null)
                        {
                            System.err.println(formatTime() + " put " + apname 
                                + " on waiting for " + clientmac);
                        }
                    }
                    return false;
                }
            }
        }
        return true;
    }

    /** generate a polling task and hand over to thread pool. */
    private void addPollTask(String apname, String clientmac, long timestamp)
    {
        // now ok to poll
        synchronized (_lock)
        {
            _working.add(apname);
            _waiting.remove(apname);
            _polled.put(apname, new Long(System.currentTimeMillis()));
        }
        try
        {
            // generate a polling task and hand over to thread pool
            SnmpTask task = new SnmpTask(this, apname, clientmac, timestamp);
            _pool.execute(task);
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }

    /** Actually poll the access point using a script. */
    private void doPoll(String apname, String clientmac, long timestamp)
        throws Exception
    {
        synchronized (_lock)
        {
            _working.add(apname);
            _waiting.remove(apname);
            _polled.put(apname, new Long(System.currentTimeMillis()));
        }
        // check if this poll fresh or a delayed request
        if (clientmac == null)
        {
            System.err.println(formatTime() + " poll " 
                + apname + " from waiting/apall list");
        }
        else
        {
            System.err.println(formatTime() + " poll " + apname + " for " 
                + clientmac + " associated at " + formatTime(timestamp));
        }
        // execute the polling script
        Process p = SystemUtil.command(
            _script + " " + apname + " " + _community);
        p.waitFor();
        // read the output of the script
        String uptime = null;
        int numClients = 0;
        BufferedReader in = new BufferedReader(
            new FileReader("/tmp/snmp/" + apname + ".snmp"));
        String line = null;
        while ((line=in.readLine()) != null)
        {
            // sys table
            if (line.startsWith("sys,"))
            {
                int x = line.lastIndexOf(",");
                uptime = line.substring(x+1);
            }
            // client table
            if (line.startsWith("c2,"))
            {
                String[] data = line.split(",");
                clientmac = convertMac(data[3]);
                String ipaddress = data[7];
                if (!ipaddress.startsWith("0.0.") 
                    && !ipaddress.equals("192.168.0.1"))
                {
                    numClients ++;
                    record(clientmac, ipaddress, apname);
                }
//                System.err.println("\t"+apname+": grabbed " + ipaddress 
//                    + " for " + clientmac);
            }
        }
        in.close();
        // update all records
        synchronized (_lock)
        {
            _waiting.remove(apname);
            _working.remove(apname);
            _apstat.put(apname, new ApStat(apname, uptime, numClients));
            _polled.put(apname, new Long(System.currentTimeMillis()));
            if (numClients == 0)
            {
                _aplist.remove(apname);
            }
            else
            {
                _aplist.add(apname);
            }
        }
    }

    /** Convert mac address from snmp format to syslog format. */
    private String convertMac(String mac)
    {
        StringBuffer ret = new StringBuffer();
        String[] data = mac.split(":");
        for (int i = 0; i < data.length; i++)
        {
            String tmp = data[i];
            if (tmp.length() == 1)
            {
                ret.append("0" + tmp);
            }
            else
            {
                ret.append(tmp);
            }
        }
        return ret.toString();
    }

    /** Record location mapping in tables. */
    private synchronized void record(
        String clientmac, String ipaddress, String apname)
    {
        _ipLocator.put(ipaddress, apname);
        _macLocator.put(clientmac, apname);
    }

    /** Format current time for logging. */
    private String formatTime()
    {
        return formatTime(System.currentTimeMillis());
    }

    private String formatTime(long timestamp)
    {
        Calendar cal = Calendar.getInstance();
        cal.setTimeInMillis(timestamp);
        int hour = cal.get(Calendar.HOUR_OF_DAY);
        int minute = cal.get(Calendar.MINUTE);
        int second = cal.get(Calendar.SECOND);
        String h = (hour<10) ? "0" + Integer.toString(hour) : Integer.toString(hour);
        String m = (minute<10) ? "0" + Integer.toString(minute) : Integer.toString(minute);
        String s = (second<10) ? "0" + Integer.toString(second) : Integer.toString(second);
        return "[" + h + ":" + m + ":" + s + "]";
    }

    //=======================================================================
    // Implementation of Runnable interface
    //=======================================================================

    /** Periodically poll the APs on the waiting list. */
    public void run()
    {
        while (true)
        {
            try
            {
                Thread.sleep(WAIT_INTERVAL);
            }
            catch (InterruptedException ie)
            {
                break;
            }
            // poll all APs on waiting list if they are pollable
            Object[] obj = null;
            synchronized (_lock)
            {
                obj = _waiting.toArray();
//                System.err.println(formatTime() 
//                    + " process waiting list - " + Arrays.asList(obj));
            }
            for (int i=0; i<obj.length; i++)
            {
                String apname = (String) obj[i];
                // ok to poll now?
                if (toPoll(apname, null))
                {
                    addPollTask(apname, null, 0);
                }
            }
            // add one more task randomly picked from AP list
            synchronized (_lock)
            {
                obj = _aplist.toArray();
//                System.err.println(formatTime() 
//                    + " process apall list - (" + obj.length + ")");
            }
            int len = obj.length;
            if (len > 0)
            {
                String apname = (String) obj[_random.nextInt(len)];
                // ok to poll now?
                if (toPoll(apname, null))
                {
                    addPollTask(apname, null, 0);
                }
            }
        }
        System.err.println("SnmpLocator: waiting list thread ended.");
    }

    //=======================================================================
    // Test
    //=======================================================================

    public static void main(String[] args)
        throws Exception
    {
        SnmpLocator locator = new SnmpLocator(args[0]);
        
        BufferedReader in = 
            new BufferedReader(new InputStreamReader(System.in), 1);
        String line = in.readLine();
        while (line != null)
        {
            locator.update(line);
            line = in.readLine();
        }
    }
}
