/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package campus.sysnort;


import serp.util.Options;
import org.apache.log4j.Logger;
import org.apache.log4j.Level;
import org.apache.log4j.BasicConfigurator;
import java.io.FileInputStream;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.List;
import java.util.Vector;

import solar.api.Id;
import solar.api.Event;
import solar.api.Attribute;
import solar.impl.EventFactory;
import solar.service.SolarServices;
import solar.service.ServiceManager;
import solar.service.dispatch.DispatchService;
import solar.service.pack.PackService;
import solar.service.pubsub.Publisher;
import campus.LogTracer;
import campus.syslog.SyslogConstants;


/**
 *  A Syslog source using Solar PACK service. The source can be subscribed
 *  by multiple Snort app, each may have multiple rules/plugins with 
 *  respective PACK policies.  The input is STDIN, either disregarding the
 *  temporal relation (non-stop) or respecting the message timestamp by
 *  setting command line option "-tracer true".
 *  
 *  @author     Guanling Chen
 *  @version    $Id: SysnortSensor.java,v 1.14 2003/10/07 01:13:23 glchen Exp $
 */
public class SysnortSensor 
    implements Publisher, LogTracer.TracerObserver, SyslogConstants
{
    private static Logger _log = Logger.getLogger("solar");

    protected Options _opts;
    protected boolean _tracer;
    protected PackService _pack;
    

    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public SysnortSensor(Options opts)
        throws Exception
    {
        _opts = opts;
        _tracer = opts.getBooleanProperty("tracer", false);

        ServiceManager sm = SolarServices.getInstance();
        sm.init(_opts);
        // initiate services first
        DispatchService tpt = (DispatchService) sm.getService("dispatch");
        _pack = (PackService) sm.getService("pack");
    }


    //=======================================================================
    // Public interface methods.
    //=======================================================================
        
    public void run()
        throws Exception
    {
        // register the publisher
        _pack.attachPublisher(new Id("SysnortSensor"), this);

        if (_tracer)
        {
            LogTracer tracer = new LogTracer(_opts);
            tracer.setObserver(this);
            tracer.run();
        }
        else
        {
            BufferedReader in = 
                new BufferedReader(new InputStreamReader(System.in), 1);
            // parse the input and publishes event
            String line = in.readLine();
            while(line != null)
            {
                handleMessage(line);
                line = in.readLine();
            }
        }
    }

    public void handleMessage(String msg)
        throws Exception
    {
        String[] fields = msg.split("\t");
        List attr = new Vector();
        int idx = 0;
        attr.add(new Attribute(EVT_TIMESTAMP, fields[idx++]));
        attr.add(new Attribute(EVT_YEAR, fields[idx++]));
        attr.add(new Attribute(EVT_MONTH, fields[idx++]));
        attr.add(new Attribute(EVT_DAY, fields[idx++]));
        attr.add(new Attribute(EVT_TIME, fields[idx++]));
        attr.add(new Attribute(EVT_MSGTYPE, fields[idx++]));
        attr.add(new Attribute(EVT_APNAME, fields[idx++]));
        attr.add(new Attribute(EVT_CLIENTHOST, fields[idx++]));
        attr.add(new Attribute(EVT_CLIENTMAC, fields[idx++]));
        attr.add(new Attribute(EVT_ACTION, fields[idx++]));
        attr.add(new Attribute(EVT_NOTE, fields[idx++]));
        Event evt = EventFactory.getPlainEvent(attr);

        System.out.println(evt);
        _pack.publish(new Id("SysnortSensor"), evt);
    }


    //=======================================================================
    //  Main entrance.
    //=======================================================================

    public static void main(String[] args)
        throws Exception
    {
        Options opts = new Options();
        opts.setFromCmdLine(args);

        // set the logging verbose level
        boolean debug = opts.getBooleanProperty("debug", false);
        if (debug)
        {
            _log.setLevel( (Level) Level.DEBUG );
        }
        else
        {
            _log.setLevel( (Level) Level.INFO );
        }
        BasicConfigurator.configure();

        // load configuration options
        String cfg = opts.getProperty("cfg");
        if (cfg != null)
            opts.load(new FileInputStream(cfg));
        
        new SysnortSensor(opts).run();
    }

}
