/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package campus.sysnort.trigger;


import serp.util.Options;
import java.util.Date;
import java.util.Set;
import java.util.HashSet;
import java.util.Hashtable;

import campus.syslog.SyslogConstants;
import campus.sysnort.Trigger;
import solar.api.Event;
import solar.util.EmailUtil;


/**
 *  A Sysnort trigger that detects a flickering card, which is associated
 *  with an AP for a short time before next rotation.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: FlickerTrigger.java,v 1.7 2003/09/08 23:50:35 glchen Exp $
 */
public class FlickerTrigger 
    extends Trigger
    implements SyslogConstants
{
    public static int HITS_THRESHOLD     = 3;   // alert if hit apset 3 times
    public static long TIME_THRESHOLD   = 60;  // 1 minute "rotation" threshold

    protected String _smtp;
    protected String _emailto;
    
    class ClientState
    {
        String _mac;
        Set _apset = new HashSet(); // the set of ap hit by card while flickering
        long _ptime = 0;    // last time of associated message
        int _hits = 0;      // number of times hitting the apset
        
        public ClientState(String mac) { _mac = mac; }
        public void clear()
        {
            _apset.clear(); _ptime = 0; _hits = 0;
        }
        public ClientState makeCopy()
        {
            ClientState cs = new ClientState(this._mac);
            cs._apset = new HashSet(this._apset);
            cs._ptime = this._ptime;
            cs._hits = this._hits;
            return cs;
        }
        public String toString()
        {
            return _mac + " " + new Date(_ptime*1000) + " " + _apset;
        }
    }
    protected Hashtable _state; // (mac, clientstate)
    protected Set _reported;    // already reported clientstate
    
    //=======================================================================
    // Implementation of campus.sysnort.Trigger interface.
    //=======================================================================

    public void init(Options opts)
        throws Exception
    {
        _smtp = opts.getProperty("smtp", "mail.cs.dartmouth.edu");
        _emailto = opts.getProperty("emailto", "glchen@cs.dartmouth.edu");
        _state = new Hashtable();
        _reported = new HashSet();
    }

    // basic strategy is to maintain a set of APs which the client associated
    // continously but stayed less than 1 minute. if the client has hit this
    // group more than 3 times, we trigger the action
    public boolean checkCondition(Event evt)
        throws Exception
    {
        boolean ret = false;
        // retrieve basic fields
        long timestamp = Long.parseLong((String)evt.getValue(EVT_TIMESTAMP));
        String msgtype = (String) evt.getValue(EVT_MSGTYPE);
        String apname = (String) evt.getValue(EVT_APNAME);
        String clienthost = (String) evt.getValue(EVT_CLIENTHOST);
        String clientmac = (String) evt.getValue(EVT_CLIENTMAC);
        String action = (String) evt.getValue(EVT_ACTION);
        // immediately filter out unwanted events
        if ((!msgtype.equals(TAG_INFO))
            || clienthost.endsWith("-ap")
            || clienthost.endsWith("-wb")
            || (!action.equals(TAG_ASSOCIATED)))
            return false;

        // update client state, first check whether current state is clean
        ClientState cs = (ClientState) _state.get(clientmac);
        if (cs == null || cs._ptime == 0)
        {
            // first time on network or came from a clean state
            cs = new ClientState(clientmac);
            cs._apset.add(apname);
            cs._ptime = timestamp;
            _state.put(clientmac, cs);
            return false;
        }

        // then we check the interval of this association to previous one
        long delta = timestamp - cs._ptime;
        if (delta > TIME_THRESHOLD)
        {
            // longer than threshold, consider normal, clear state
            cs.clear();
            cs._apset.add(apname);
            cs._ptime = timestamp;
        }
        else
        {
            // otherwise update the apset and hits
            if (cs._apset.contains(apname)) { cs._hits ++; }
            cs._apset.add(apname);
            cs._ptime = timestamp;
            if (cs._hits > HITS_THRESHOLD)
            {
                ret = true;
                _actionObject = cs.makeCopy();
                cs.clear();
            }
        }
        return ret;
    }

    public void takeAction(Object actionObject)
        throws Exception
    {
        ClientState cs = (ClientState) actionObject;
        String reportKey = cs._mac + " " + cs._apset.toString();
        // has this apset/client been reported before?
        if (_reported.contains(reportKey))
        {
            return;
        }
        else
        {
            _reported.add(reportKey);
        }

        System.out.println(this.getName() + ": " + cs);

        /*
        String msg = "Around time " + new Date(cs._ptime*1000) + " the card " 
            + cs._mac + " is suspected to flicker "
            + "among these access points: \n\n" + cs._apset
            + "\n\n";

        EmailUtil.sendmail(
            _smtp, 
            _emailto,
            "[SYSNORT] Card Flickering Alert",
            msg);*/
    }
}
