/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.module.protocol;


import java.util.Collection;
import java.util.LinkedList;

import solar.api.Event;


/**
 *  A buffer contains an individual queue for one receiver or an aggregated
 *  queue for multiple receivers.  When the queue reaches the limit, it either
 *  drop the new event (drop_tail) or the oldest one (drop_head).
 *  
 *  @author     Guanling Chen
 *  @version    $Id: Buffer.java,v 1.5 2003/10/13 13:06:56 glchen Exp $
 */
public class Buffer 
{
    protected int _limit = Integer.MAX_VALUE;
    protected boolean _drophead;
    protected LinkedList _buffer;

    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public Buffer()
    {
        _buffer = new LinkedList();
    }

    public Buffer(int limit)
    {
        _limit = limit;
        _buffer = new LinkedList();
    }

    public Buffer(Collection collection)
    {
        _buffer = new LinkedList(collection);
    }

    //=======================================================================
    // Public interface methods.
    //=======================================================================

    /** Add an event to the buffer, return false if limit is exceeded. */
    public synchronized void put(Event evt)
        throws BufferFullException
    {
        if (size() < _limit)
        {
            _buffer.addLast(evt);
        }
        else if (_drophead)
        {
            _buffer.removeFirst();
            _buffer.addLast(evt);
        }
        else
        {
            // just drop on floor
            return;
        }
    }

    /** Get and remove an event from the buffer. */
    public synchronized Event get()
    {
        return (Event) _buffer.removeFirst();
    }

    /** Get current size of the buffer. */
    public synchronized int size()
    {
        return _buffer.size();
    }

    /** Clear out the whole buffer. */
    public synchronized void clear()
    {
        _buffer.clear();
    }

    public synchronized void dropTail()
    {
        _buffer.removeLast();
    }

    public synchronized void dropHead()
    {
        _buffer.removeFirst();
    }

    public synchronized Collection getAll()
    {
        return _buffer;
    }

    public synchronized void putAll(Collection collection)
    {
        _buffer.addAll(collection);
    }
}
