/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */
 
package solar.service.directory;


import serp.util.Options;
import org.apache.log4j.Logger;
import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.util.Vector;
import java.util.Iterator;
import java.util.Hashtable;

import solar.api.Id;
import solar.api.Request;
import solar.api.Address;
import solar.service.BaseService;
import solar.service.ServiceException;
import solar.service.directory.comm.*;
import solar.service.rpc.RpcService;
import solar.service.rpc.RpcConstants;
import solar.service.rpc.RpcException;
import solar.service.transport.TransportService;
import solar.service.transport.TransportException;


/**
 *  A distributed implementation of directory service.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: DistDirectoryService.java,v 1.8 2003/10/05 18:50:56 glchen Exp $
 */
public class DistDirectoryService 
    extends BaseService
    implements DirectoryService, DirectoryConstants, RpcConstants
{
    /** The logger for distributed directory service. */
    protected Logger _log = Logger.getLogger(DistDirectoryService.class);

    protected Address _laddr;
    protected RpcService _rpc;
    protected TransportService _transport;
    // instance of name space if this node is the server
    protected Directory _directory;
    // list of listeners on this node while the server is a remote node
    protected Hashtable _listeners;
    
    
    //=======================================================================
    // Implementation of solar.service.Service interface.
    //=======================================================================
    
    public void initDependents()
        throws ServiceException
    {
        super.initDependents();     // must call at the top
        
        try
        {
            // get transport service and local address
            _transport = (TransportService) 
                getDependentService("transport", "transport");
            _laddr = _transport.getLocalAddress();
            // get RPC service
            _rpc = (RpcService) getDependentService("rpc", "rpc");
            // test if we have a DHT-capable transport service
            _transport.getAddressByKey("test_test");
            
        }
        catch (TransportException te)
        {
            throw new ServiceException(
                "Failed to get a DHT-capable transport service.", te);
        }
        _directory = new FlatDirectory();
        _listeners = new Hashtable();
    }

    public Object handleRequest(Request request)
      throws Exception
    {
        Object ret = null;
        if (request instanceof AddListenerRequest)
        {
            AddListenerRequest alr = (AddListenerRequest) request;
            Name pattern = alr.getPattern();
            RemoteDirectoryListener listener = 
                (RemoteDirectoryListener) alr.getListener();
            if (_laddr.equals(listener.getAddress()))
            {
                // turns out to be a local listener
                DirectoryListener dob = 
                    (DirectoryListener) _listeners.get(listener.getId());
                if (dob != null)
                {
                    _directory.addListener(pattern, dob);
                }
            }
            else
            {
                // indeed a remote listener
                listener.setLogger(_log);
                listener.setRpcService(getName(), _rpc);
                _directory.addListener(pattern, listener);
            }
        }
        else if (request instanceof RemoveListenerRequest)
        {
            RemoveListenerRequest rlr = (RemoveListenerRequest) request;
            Name pattern = rlr.getPattern();
            RemoteDirectoryListener listener = 
                (RemoteDirectoryListener) rlr.getListener();
            _directory.removeListener(pattern, listener);
        }
        else if (request instanceof LookupRequest)
        {
            LookupRequest lr = (LookupRequest) request;
            Name pattern = lr.getPattern();
            ret = _directory.lookup(pattern);
        }
        else if (request instanceof AdvertiseRequest)
        {
            AdvertiseRequest ar = (AdvertiseRequest) request;
            Name name = ar.getName();
            NameRecord rec = ar.getNameRecord();
            NameRecord or = _directory.advertise(name, rec);
            ret = (or == null) ? null : or.getObject();
        }
        else if (request instanceof UnadvertiseRequest)
        {
            UnadvertiseRequest ur = (UnadvertiseRequest) request;
            Name name = ur.getName();
            NameRecord rec = ur.getNameRecord();
            NameRecord or = _directory.unadvertise(name, rec);
            ret = (or == null) ? null : or.getObject();
        }
        else if (request instanceof UpdateRequest)
        {
            UpdateRequest ur = (UpdateRequest) request;
            Name oldname = ur.getOldName();
            NameRecord oldrec = ur.getOldNameRecord();
            Name newname = ur.getNewName();
            NameRecord newrec = ur.getNewNameRecord();
            NameRecord or = _directory.update(oldname, oldrec, newname, newrec);
            ret = (or == null) ? null : or.getObject();
        }
        else if (request instanceof NameAddedRequest)
        {
            NameAddedRequest nar = (NameAddedRequest) request;
            Id id = nar.getId();
            Name name = nar.getName();
            NameRecord rec = nar.getNameRecord();
            DirectoryListener listener = 
                (DirectoryListener) _listeners.get(id);
            if (listener != null) listener.nameAdded(name, rec);
            else throw new DirectoryException("Name listener not found.");
        }
        else if (request instanceof NameRemovedRequest)
        {
            NameRemovedRequest nrr = (NameRemovedRequest) request;
            Id id = nrr.getId();
            Name name = nrr.getName();
            NameRecord rec = nrr.getNameRecord();
            DirectoryListener listener = 
                (DirectoryListener) _listeners.get(id);
            if (listener != null) listener.nameRemoved(name, rec);
            else throw new DirectoryException("Name listener not found.");
        }
        else if (request instanceof NameUpdatedRequest)
        {
            NameUpdatedRequest nur = (NameUpdatedRequest) request;
            Id id = nur.getId();
            Name oldname = nur.getOldName();
            Name newname = nur.getNewName();
            NameRecord oldrec = nur.getOldNameRecord();
            NameRecord newrec = nur.getNewNameRecord();
            DirectoryListener listener = 
                (DirectoryListener) _listeners.get(id);
            if (listener != null) listener.nameUpdated(oldname, oldrec, newname, newrec);
            else throw new DirectoryException("Name listener not found.");
        }
        else if (request instanceof NameMigratedRequest)
        {
            NameMigratedRequest nmr = (NameMigratedRequest) request;
            Id id = nmr.getId();
            Name name = nmr.getName();
            NameRecord oldrec = nmr.getOldNameRecord();
            NameRecord newrec = nmr.getNewNameRecord();
            DirectoryListener listener = 
                (DirectoryListener) _listeners.get(id);
            if (listener != null) listener.nameMigrated(name, oldrec, newrec);
            else throw new DirectoryException("Name listener not found.");
        }
        else
        {
            throw new RpcException(
                "Failed to handle directory request " + request + ".");
        }
        return ret;
    }

    //=======================================================================
    // Implementation of solar.service.directory.DirectoryService interface.
    //=======================================================================
    
    public void addListener(Name pattern, DirectoryListener listener)
        throws DirectoryException
    {
        // add listener to locally maintained table
        Id id = Id.makeId("directory-listener-");
        _listeners.put(id, listener);

        // create a remote proxy and tell server to add it
        RemoteDirectoryListener rno = 
            new RemoteDirectoryListener(id, _laddr);
            
        Request req = new AddListenerRequest(pattern, rno);
        requestAll(pattern, req);
    }
        

    public void removeListener(Name pattern, DirectoryListener listener)
        throws DirectoryException
    {
        // remove listener from locally maintained table
        Id oid = null;
        Iterator itr = _listeners.keySet().iterator();
        while (itr.hasNext())
        {
            Id id = (Id) itr.next();
            if (_listeners.get(id) == listener) // reference comparison
            {
                oid = id;
                break;
            }
        }
        if (oid != null)
        {
            _listeners.remove(oid);
        }
        else
        {
            throw new DirectoryException(
                "The listener to remove does not exist.");
        }
            
        // create a remote proxy and tell server to remove it
        RemoteDirectoryListener rno = 
            new RemoteDirectoryListener(oid, _laddr);
            
        Request req = new RemoveListenerRequest(pattern, rno);
        requestAll(pattern, req);
    }

    public List lookup(Name pattern)
        throws DirectoryException
    {
        Set ret = new HashSet();
        Request req = new LookupRequest(pattern);
        Iterator itr = requestAll(pattern, req).iterator();
        while (itr.hasNext())
        {
            ret.addAll((List)itr.next());
        }
        return new Vector(ret);
    }

    public Object advertise(Name name, Id guid, Object obj)
        throws DirectoryException
    {
        Request req = new AdvertiseRequest(
            name, new NameRecord(guid, _laddr, obj));
        requestAll(name, req);
        return null;
    }

    public Object unadvertise(Name name, Id guid)
        throws DirectoryException
    {
        Request req = new UnadvertiseRequest(
            name, new NameRecord(guid, _laddr));
        requestAll(name, req);
        return null;
    }

    public Object update(Name oldname, Id guid, Name newname, Object obj)
        throws DirectoryException
    {
        unadvertise(oldname, guid);
        advertise(newname, guid, obj);
        return null;
    }


    //=======================================================================
    // Private helper methods.
    //=======================================================================

    private List requestAll(Name name, Request req)
        throws DirectoryException
    {
        List ret = new Vector();
        Iterator itr = name.getStrands().iterator();
        try
        {
            while (itr.hasNext())
            {
                String strand = (String) itr.next();
                Address dest = _transport.getAddressByKey(strand);
                Object obj = _rpc.invoke(
                    dest, getName(), req, RPC_TIMEOUT);
                if (obj != null) ret.add(obj);
            }
        }
        catch (Exception ex)
        {
            throw new DirectoryException(ex);
        }
        return ret;
    }
}
