/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.fap;


import solar.api.Id;
import solar.api.Identifiable;


/**
 *   A pipe consists of a pair of channels: A writable sink channel and a 
 *   readable source channel. A pipe is unidirectional.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: Pipe.java,v 1.12 2003/09/25 05:30:02 glchen Exp $
 */
public abstract class Pipe 
    implements Identifiable
{
    protected Id _pid;
    protected int _type;
    protected SourceChannel _source;
    protected SinkChannel _sink;
    
    //=======================================================================
    // Constructors.
    //=======================================================================

    /** Constructs a pipe with specified ID and pull/push type. */
    public Pipe(Id pid, int type)
    {
        _pid = pid;
        _type = type;
    }

    /** Constructs a pipe with additional pair of source/sink channels. */
    public Pipe(Id pid, int type, 
        SourceChannel source, SinkChannel sink)
    {
        _pid = pid;
        _type = type;
        _source = source;
        _sink = sink;
    }

    
    //=======================================================================
    // Implementation of solar.api.Identifiable interface.
    //=======================================================================

    public void setId(Id id)
    {
        throw new IllegalStateException("The pipe ID has already been set.");
    }
    
    public Id getId()
    {
        return _pid;
    }


    //=======================================================================
    // Public interface methods.
    //=======================================================================

    /** Return whether this is a push or pull pipe. */
    public int getType()
    {
        return _type;
    }

    /** Attach the source channel to this pipe. */
    public void attachSourceChannel(SourceChannel source)
        throws FapException
    {
        if (_source != null)
        {
            throw new FapException("Source channel already attached.");
        }
        _source = source;
    }

    /** Return the source channel of this pipe. */
    public SourceChannel getSourceChannel()
    {
        return _source;
    }

    /** Return the name of the input port for this pipe. */
    public String getInputPort()
        throws FapException
    {
        return _source.getOutputPort().getName();
    }

    /** Attach a sink channel to this pipe. */
    public void attachSinkChannel(SinkChannel sink)
        throws FapException
    {
        if (_sink != null)
        {
            throw new FapException("Sink channel already attached.");
        }
        _sink = sink;
    }

    /** Return the sink channel of this pipe. */
    public SinkChannel getSinkChannel()
    {
        return _sink;
    }

    /** Return the name of the output port for this pipe. */
    public String getOutputPort()
        throws FapException
    {
        return _sink.getInputPort().getName();
    }

    /** Push data through this pipe. */
    public abstract void pushData(Object data)
        throws FapException;

    /** Pull data from this pipe. */
    public abstract Object pullData(Object query)
        throws FapException;
}
