/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.mhost;


import org.apache.log4j.Logger;
import serp.util.Options;
import java.util.List;
import java.util.Vector;
import java.util.Iterator;

import solar.api.Request;
import solar.api.Address;
import solar.service.BaseService;
import solar.service.ServiceException;
import solar.service.dispatch.DispatchService;
import solar.service.mhost.comm.AttachRequest;
import solar.service.mhost.comm.DetachRequest;
import solar.service.rpc.RpcService;
import solar.service.rpc.RpcException;
import solar.service.rpc.RpcConstants;


/**
 *  Solar implementation of mobile host service.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: SolarMobileHostService.java,v 1.2 2003/10/05 18:50:57 glchen Exp $
 */
public class SolarMobileHostService 
    extends BaseService
    implements MobileHostService, RpcConstants
{
    /** Logging facility. */
    protected static Logger _log = 
        Logger.getLogger(SolarMobileHostService.class);

    protected boolean _server;
    protected String _mclientName;
    protected String _transportName;
    protected Address _transportAddress;

    protected RpcService _rpc;
    protected DispatchService _dispatch;

    // if _server is false, then this list has only one-element - the address
    // of the attached server; otherwise this list contains all currently
    // attached clients
    protected List _attached = new Vector();
    protected List _listeners = new Vector();


    //=======================================================================
    //  Implementation of solar.service.Service interface.
    //=======================================================================

    public void init()
        throws ServiceException
    {
        _mclientName = getName();
        Options config = getServiceManager().getServiceConfig(_mclientName);
        if (config.getProperty("server") == null)
        {
            throw new IllegalArgumentException(
                "Option 'server' not specified for mobile host service.");
        }
        _server = config.getBooleanProperty("server");
        _transportName = config.getProperty("transport", "transport");
    }
    
    public void initDependents()
        throws ServiceException
    {
        super.initDependents();     // must call at the top

        _dispatch = (DispatchService) 
            getDependentService("dispatch", "dispatch");
        _transportAddress = 
            _dispatch.getTransportAddress(_transportName);
        _rpc = (RpcService) 
            getDependentService("rpc", "rpc");
    }

    //=======================================================================
    //  Implementation of MobileHostService interface.
    //=======================================================================

    public void attach(Address address)
        throws MobileHostException
    {
        Request ar = null;
        if (_server)
        {
            if (isAttached(address))
            {
                throw new IllegalStateException(
                    "The client has already been attached.");
            }
            ar = new AttachRequest(address, _transportAddress);
        }
        else
        {
            if (isAttached())
            {
                throw new IllegalStateException(
                    "The client has already been attached.");                
            }
            ar = new AttachRequest(_transportAddress, address);
        }
        try
        {
            callRpc(address, ar);
        }
        catch (RpcException re)
        {
            throw new MobileHostException(re);
        }
        synchronized (_attached)
        {
            _attached.add(address);
            notifyClientAttached(address);
        }
    }

    public void detach(Address address)
        throws MobileHostException
    {
        Request dr = null;
        if (_server)
        {
            if (!isAttached(address))
            {
                throw new IllegalStateException(
                    "The client is not attached yet.");
            }
            dr = new DetachRequest(address, _transportAddress);
        }
        else
        {
            if (!isAttached())
            {
                throw new IllegalStateException(
                    "The client is not attached yet.");                
            }
            dr = new DetachRequest(_transportAddress, address);
        }
        try
        {
            callRpc(address, dr);
        }
        catch (RpcException re)
        {
            throw new MobileHostException(re);
        }
        synchronized (_attached)
        {
            _attached.add(address);
            notifyClientDetached(address);
        }
    }

    public boolean isAttached()
    {
        if (_server)
        {
            throw new IllegalStateException(
                "Cannot call isAttached on a server.");
        }
        synchronized (_attached)
        {
            return _attached.size() != 0;
        }
    }

    public boolean isAttached(Address client)
    {
        if (!_server)
        {
            throw new IllegalStateException(
                "Cannot call isAttached(Address) on a client.");
        }
        synchronized (_attached)
        {
            return _attached.contains(client);
        }
    }

    public Address getAttachedServer()
    {
        if (_server)
        {
            throw new IllegalStateException(
                "Cannot call getAttachedServer on a server.");
        }
        synchronized (_attached)
        {
            return (Address) _attached.get(0);
        }
    }

    public List getAttachedClients()
    {
        if (!_server)
        {
            throw new IllegalStateException(
                "Cannot call getAttachedClients on a client.");
        }
        return _attached;        
    }
    
    public void addListener(MobileHostListener listener)
    {
        synchronized (_listeners)
        {
            _listeners.add(listener);
        }
    }

    public void removeListener(MobileHostListener listener)
    {
        synchronized (_listeners)
        {
            _listeners.remove(listener);
        }        
    }

    //=======================================================================
    //  Private helper methods.
    //=======================================================================
    
    private Object callRpc(Address address, Request request)
        throws RpcException
    {
        return _rpc.invoke(address, _mclientName, request, RPC_TIMEOUT);
    }

    private void notifyClientAttached(Address address)
    {
        synchronized (_listeners)
        {
            Iterator itr = _listeners.iterator();
            while (itr.hasNext())
            {
                ((MobileHostListener) itr.next()).clientAttached(address);
            }
        }
    }

    private void notifyClientDetached(Address address)
    {
        synchronized (_listeners)
        {
            Iterator itr = _listeners.iterator();
            while (itr.hasNext())
            {
                ((MobileHostListener) itr.next()).clientDetached(address);
            }
        }
    }
}
