/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.pack;


import serp.util.Options;
import org.apache.log4j.Logger;
import java.util.Iterator;

import solar.api.Id;
import solar.api.Guid;
import solar.api.Event;
import solar.api.Address;
import solar.api.Request;
import solar.service.BaseService;
import solar.service.ServiceException;
import solar.service.dispatch.DispatchService;
import solar.service.mhost.MobileHostService;
import solar.service.mhost.MobileHostListener;
import solar.service.pack.proto.BufferedEvent;
import solar.service.pack.proto.InputProtocol;
import solar.service.pack.proto.OutputProtocol;
import solar.service.pack.comm.PackEvent;
import solar.service.pack.comm.SubscribeEvent;
import solar.service.pack.comm.SubscribeRequest;
import solar.service.pack.comm.AttachPublisherRequest;
import solar.service.pack.comm.AttachSubscriberRequest;
import solar.service.pack.comm.DetachPublisherRequest;
import solar.service.pack.comm.DetachSubscriberRequest;
import solar.service.pubsub.Publisher;
import solar.service.pubsub.Subscriber;
import solar.service.rpc.RpcService;
import solar.service.rpc.RpcConstants;
import solar.service.rpc.RpcException;
import solar.service.transport.SocketAddress;


/**
 *  The PACK service that runs at the client side.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: ClientPackService.java,v 1.21 2003/10/13 03:13:03 glchen Exp $
 */
public class ClientPackService 
    extends BaseService
    implements PackService, MobileHostListener, RpcConstants, PackConstants
{
    /** Logging facility. */
    protected static Logger _log = 
        Logger.getLogger(ClientPackService.class);

    protected Address _server;
    protected String _packName;
    protected String _transportName;
    protected Address _transportAddress;

    protected RpcService _rpc;
    protected MobileHostService _mhost;
    protected DispatchService _dispatch;

    protected InputProtocol _inproto;
    protected OutputProtocol _outproto;
    protected EndPointManager _epman = new EndPointManager();
    
    //=======================================================================
    // Implementation of solar.service.Service interface.
    //=======================================================================

    public void init()
        throws ServiceException
    {
        _packName = getName();
        Options config = getServiceManager().getServiceConfig(_packName);
        _transportName = config.getProperty("transport", "transport");
        String server = config.getProperty("server");
        if (server != null)
        {
            try {
                _server = new SocketAddress(server);
            }
            catch (Exception ex) {
                throw new ServiceException(ex);
            }
        }
        int bt = 
            config.getIntProperty("buffer_limit", PACK_BUFFER_LIMIT);
        int mt = 
            config.getIntProperty("memory_limit", PACK_MEMORY_LIMIT);
        int st = 
            config.getIntProperty("suspend_limit", PACK_SUSPEND_LIMIT);
            
        _inproto = new InputProtocol(bt, mt, st);
        _outproto = new OutputProtocol(bt, mt, st);
        new Thread(new Sender(), "pack-sender").start();
        new Thread(new Handler(), "pack-handler").start();
    }
    
    public void initDependents()
        throws ServiceException
    {
        super.initDependents();     // must call at the top

        // retrieve needed services
        _dispatch = (DispatchService) 
            getDependentService("dispatch", "dispatch");
        _transportAddress = 
            _dispatch.getTransportAddress(_transportName);
        _rpc = (RpcService) 
            getDependentService("rpc", "rpc");
        _mhost = (MobileHostService) 
            getDependentService("mhost", "mhost");

        // attach to specified server (Planet) and register listener
        if (_server != null)
        {
            _mhost.attach(_server);
        }
        _mhost.addListener(this);
    }

    public void handleEvent(Event evt)
        throws Exception
    {
        // special trick if we use Transport instead of ObjectTransport
        if (PackEvent.isPackEvent(evt)) {
            PackEvent pe = (evt instanceof PackEvent) ? 
                (PackEvent) evt : new PackEvent(evt);
            handlePackEvent(pe);
        }
        else if (SubscribeEvent.isSubscribeEvent(evt)) {
            SubscribeEvent se = (evt instanceof SubscribeEvent) ? 
                (SubscribeEvent) evt : new SubscribeEvent(evt);
            handleSubscribeEvent(se);
        }
        else {
            throw new IllegalArgumentException(
                "Unexpected event received: " + evt);
        }
    }

    public void handleSubscribeEvent(SubscribeEvent evt)
        throws Exception
    {
        Id pubId = evt.getPubId();
        Id subId = evt.getSubId();
        Object policy = evt.getPolicy();
        _outproto.addPublisher(pubId);
        _outproto.addSubscriber(pubId, subId, policy, _server);
    }
    
    public void handlePackEvent(PackEvent evt)
        throws Exception
    {
        // handle events for local subscribers
        Id pubId = evt.getPubId();
        _inproto.putEvent(evt, null);
    }

    //=======================================================================
    // Implementation of PackService interface.
    //=======================================================================
    
    public void attachPublisher(Id pubId, Publisher publisher)
        throws PackException
    {
        assertAttached();
        _epman.assertNotAttached(pubId);
        
        try
        {
            // create a request and ask proxy for a session ID
            AttachPublisherRequest apr = new AttachPublisherRequest(pubId);
            apr.setSourceAddress(_transportAddress);
            Id session = (Id) callRpc(apr);
            // register the end point
            _epman.attach(pubId, session, _transportAddress, publisher);
        }
        catch (RpcException re)
        {
            throw new PackException(re);
        }
    }

    public void detachPublisher(Id pubId)
        throws PackException
    {
        Request dpr = new DetachPublisherRequest(pubId);
        try
        {
            callRpc(dpr);
            _epman.detach(pubId);
        }
        catch (RpcException re)
        {
            throw new PackException(re);
        }   
    }

    public void attachSubscriber(Id subId, Subscriber subscriber)
        throws PackException
    {
        assertAttached();
        _epman.assertNotAttached(subId);
        
        try
        {
            // create a request and ask proxy for a session ID
            AttachSubscriberRequest asr = new AttachSubscriberRequest(subId);
            asr.setSourceAddress(_transportAddress);
            Id session = (Id) callRpc(asr);
            // register the end point
            _epman.attach(subId, session, _transportAddress, subscriber);
        }
        catch (RpcException re)
        {
            throw new PackException(re);
        }
    }

    public void detachSubscriber(Id subId)
        throws PackException
    {
        Request dsr = new DetachSubscriberRequest(subId);
        try
        {
            callRpc(dsr);
            _epman.detach(subId);
        }
        catch (RpcException re)
        {
            throw new PackException(re);
        }   
    }
    
    public void subscribe(Id pubId, Id subId, Object policy)
        throws PackException
    {
        assertAttached();
        _epman.assertAttached(subId);
        
        try
        {
            // create a request and ask proxy for subscription
            SubscribeRequest sr = new SubscribeRequest(pubId, subId, policy);
            sr.setSourceAddress(_transportAddress);
            callRpc(sr);
            // allocate inbound buffer to receive subscribed events
            //_inproto.addPublisher(pubId);
            //_inproto.addSubscriber(pubId, subId, policy);
            Id fakeId = ScribePackService.makeFakeId(pubId, subId);
            _inproto.addPublisher(fakeId);
            _inproto.addSubscriber(fakeId, subId, policy);
        }
        catch (RpcException re)
        {
            throw new PackException(re);
        }        
    }

    public void publish(Id pubId, Event evt)
        throws PackException
    {
        assertAttached();
        _epman.assertAttached(pubId);
        
        if (_outproto.hasPublisher(pubId))
        {
            _outproto.putEvent(new PackEvent(pubId, null, evt, false), null);
        }
    }

    public Guid makeGuid(Id id)
    {
        return new Guid(_transportAddress, id);
    }
    
    //=======================================================================
    // Private helper methods.
    //=======================================================================

    private void assertAttached()
        throws PackException
    {
        if (!_mhost.isAttached())
        {
            throw new PackException("The client has not attached to server.");
        }
    }
    
    private Object callRpc(Request request)
        throws RpcException
    {
        return _rpc.invoke(_server, _packName, request, RPC_TIMEOUT);
    }

    //=======================================================================
    // Implementation of solar.service.mhost.MobileHostListener interface.
    //=======================================================================

    public void clientAttached(Address address)
    {
    }

    public void clientDetached(Address address)
    {
    }

    //=======================================================================
    // Inner helper classes.
    //=======================================================================
    
    class Sender implements Runnable
    {
        public void run()
        {
            while (true)
            {
                try
                {
                    BufferedEvent evt = (BufferedEvent) _outproto.getEvent();
                    _dispatch.dispatchEvent(
                        evt.getEvent(), _server, _packName, _transportName);
                    _log.debug("sent " + evt.getEvent());
                }
                //TODO: do something for exception, donot lose it
                catch (Exception ex)
                {
                    _log.error(ex);
                }
            }
        }
    }
    //TODO: use thread pool to handle events
    class Handler implements Runnable
    {
        public void run()
        {
            while (true)
            {
                try
                {
                    BufferedEvent evt = (BufferedEvent) _inproto.getEvent();
                    Id subId = evt.getSubId();
                    Subscriber subscriber = (Subscriber) _epman.getEndPoint(subId);
                    if (subscriber != null)
                    {
                        PackEvent pe = (PackEvent) evt.getEvent();
                        subscriber.handleEvent(pe.getEvent());
                    }
                    else
                    {
                        _log.error("Received event for non-exist subscriber " + subId);
                    }
                }
                catch (Exception ex)
                {
                    _log.error(ex);
                }
            }
        }
    }
}
