/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.pack.client;


import serp.util.Options;
import org.apache.log4j.Logger;
import java.io.FileInputStream;
import java.util.List;

import solar.api.Id;
import solar.api.Guid;
import solar.api.Event;
import solar.api.Address;
import solar.service.SolarServices;
import solar.service.ServiceManager;
import solar.service.dispatch.DispatchService;
import solar.service.mhost.MobileHostService;
import solar.service.pack.PackService;
import solar.service.pubsub.Publisher;
import solar.service.pubsub.Subscriber;
import solar.Planet;


/**
 *  The facade class to interact with Solar system for PACK service.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: PackClient.java,v 1.14 2003/10/12 00:58:38 glchen Exp $
 */
public class PackClient 
{
    protected static Logger _log = Logger.getLogger("solar");
    
    private DispatchService _dispatch;
    private PackService _pack;
    private ServiceManager _servman;

    
    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public PackClient(String host, int port)
        throws Exception
    {
        Options opts = getDefaultOptions();
        opts.setProperty("service.pack.server", 
            host + ":" + Integer.toString(port));
        construct(opts);
    }

    public PackClient(String host, int port, Options opts)
        throws Exception
    {
        String debug = opts.getProperty("debug");
        String layout = opts.getProperty("layout");
        String classname = opts.getProperty("service.edge_transport.classname");
        String hostname = opts.getProperty("service.edge_transport.hostname");
        String newport = opts.getProperty("service.edge_transport.port");
        String serialize = opts.getProperty("serialize");
        if (serialize == null)
        {
            serialize = opts.getProperty("service.edge_transport.serialize");
        }
        if (opts.getProperty("cfg") == null)
        {
            opts = getDefaultOptions();
        }
        opts.setProperty("debug", debug);
        opts.setProperty("layout", layout);
        if (classname != null)
            opts.setProperty("service.edge_transport.classname", classname);
        if (hostname != null)
            opts.setProperty("service.edge_transport.hostname", hostname);
        if (newport != null)
            opts.setProperty("service.edge_transport.port", newport);
        opts.setProperty("service.edge_transport.serialize", serialize);
        opts.setProperty("service.pack.server", 
            host + ":" + Integer.toString(port));
        construct(opts);
    }

    public PackClient(String cfg)
        throws Exception
    {
        Options opts = new Options();
        opts.load(new FileInputStream(cfg));
        construct(opts);
    }

    public PackClient(Options opts)
        throws Exception
    {
        construct(opts);
    }

    private Options getDefaultOptions()
    {
        Options opts = new Options();
        opts.setProperty("service.dispatch.classname", 
            "solar.service.dispatch.SolarDispatchService");
        opts.setProperty("service.dispatch.transport", "edge_transport");

        opts.setProperty("service.pack.classname", 
            "solar.service.pack.ClientPackService");
        opts.setProperty("service.pack.rpc", "edge_rpc");
        opts.setProperty("service.pack.transport", "edge_transport");
        opts.setProperty("service.mhost.classname", 
            "solar.service.mhost.SolarMobileHostService");
        opts.setProperty("service.mhost.rpc", "edge_rpc");
        opts.setProperty("service.mhost.transport", "edge_transport");
        opts.setProperty("service.mhost.server", "false");

        opts.setProperty("service.edge_rpc.classname",
            "solar.service.rpc.SolarRpcService");
        opts.setProperty("service.edge_rpc.transport", "edge_transport");
            
        opts.setProperty("service.edge_transport.classname", 
            "solar.service.transport.TcpTransportService");
        opts.setProperty("service.edge_transport.port", 0);
        return opts;
    }

    /** Common operations needed by constructors. */
    private void construct(Options opts)
        throws Exception
    {
        // set the logging verbose level
        Planet.configLogger(_log, opts);

        // init service manager and relavant services
        _servman = SolarServices.getInstance();
        _servman.init(opts);
        _dispatch = (DispatchService) _servman.getService("dispatch");
        _pack = (PackService) _servman.getService("pack");
    }


    //=======================================================================
    // Public interface methods.
    //=======================================================================

    public void attachPublisher(Id pubId, Publisher publisher)
        throws Exception
    {
        _pack.attachPublisher(pubId, publisher);
    }

    public void detachPublisher(Id pubId)
        throws Exception
    {
        _pack.detachPublisher(pubId);
    }

    public void attachSubscriber(Id subId, Subscriber subscriber)
        throws Exception
    {
        _pack.attachSubscriber(subId, subscriber);
    }

    public void detachSubscriber(Id subId)
        throws Exception
    {
        _pack.detachSubscriber(subId);
    }
    
    public void subscribe(Id pubId, Id subId, Object policy)
        throws Exception
    {
        _pack.subscribe(pubId, subId, policy);
    }

    public void publish(Id pubId, Event evt)
        throws Exception
    {
        _pack.publish(pubId, evt);
    }

    public Guid makeGuid(Id id)
    {
        return _pack.makeGuid(id);
    }

    public MobileHostService getMobileHostService()
        throws Exception
    {
        return (MobileHostService) _servman.getService("mhost");
    }

    
    //=======================================================================
    // Main entrance for testing purpose.
    //=======================================================================

    public static void main(String[] args)
        throws Exception
    {
        Options opts = new Options();
        opts.setFromCmdLine(args);

        String host = opts.getProperty("host", "hogback");
        int port = opts.getIntProperty("port", 5470);
        PackClient client = new PackClient(host, port);
    }
}
