/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.pack.proto;


import org.apache.log4j.Logger;
import java.util.Iterator;
import java.util.Hashtable;

import solar.api.Id;
import solar.api.Event;
import solar.api.Address;
import solar.module.protocol.Buffer;
import solar.module.protocol.Protocol;
import solar.module.protocol.ProtocolEvent;
import solar.module.protocol.BufferFullException;
import solar.service.pack.comm.PackEvent;


/**
 *  The base class for PACK protocol suit.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: BaseProtocol.java,v 1.11 2003/10/13 03:13:05 glchen Exp $
 */
public class BaseProtocol 
    implements Protocol
{
    protected static Logger _log;

    class ProtocolState
    {
        int _size;
        int _bufferLimit;
        int _memoryLimit;
        int _suspendLimit;
        Hashtable _buffers;   // (pubId, (subId/address, bufferHandle))
        public ProtocolState(int bt, int mt, int st) {
            _bufferLimit = bt;
            _memoryLimit = mt;
            _suspendLimit = st;
            _buffers = new Hashtable();
        }
    }
    
    protected ProtocolState _state;
    protected InputBufferManager _inbufman;
    protected OutputBufferManager _outbufman;

    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public BaseProtocol(int bt, int mt, int st)    
    {
        _log = Logger.getLogger(BaseProtocol.class);
        
        _state = new ProtocolState(bt, mt, st);
        _inbufman = BufferManagerFactory.makeInputBufferManager(_state);
        _outbufman = BufferManagerFactory.makeOutputBufferManager(_state);
    }

    //=======================================================================
    // Public interface methods.
    //=======================================================================

    /** Add a publisher that may put events in this buffer manager. */
    public void addPublisher(Id pubId)
    {
        synchronized (_state)
        {
            if (_state._buffers.get(pubId) == null)
            {
                _state._buffers.put(pubId, new Hashtable());
            }
        }
    }

    /** Whether this manager contains the specified publisher. */
    public boolean hasPublisher(Id pubId)
    {
        synchronized (_state)
        {
            return _state._buffers.get(pubId) != null;
        }
    }

    /** Add a local subscriber that may get events from this manager. */
    public void addSubscriber(Id pubId, Id subId, Object policy)
    {
        assertPubRegistered(pubId);
        Hashtable handles = null;
        synchronized (_state)
        {
            handles = (Hashtable) _state._buffers.get(pubId);
            BufferHandle handle = (BufferHandle) handles.get(subId);
            if (handle == null)
            {
                handle = new SingleBufferHandle(
                    pubId, subId, _state._suspendLimit);
                handles.put(subId, handle);
            }
            handle.addSubscriber(subId, policy, null);
        }
    }

    /** Add a remote subscriber that may get events from this manager. */
    public void addSubscriber(Id pubId, Id subId, Object policy, Address from)
    {
        assertPubRegistered(pubId);
        Hashtable handles = null;
        synchronized (_state)
        {
            handles = (Hashtable) _state._buffers.get(pubId);
            // try to find previous handle if there is any.
            // this is only used by subrootproto to support mobility
            Buffer buffer = null;
            Address keyaddr = null;
            BufferHandle handle = null;
            Iterator itr = handles.keySet().iterator();
            while (itr.hasNext()) {
                keyaddr = (Address) itr.next();
                handle = (BufferHandle) handles.get(keyaddr);
                // if subscriber didn't change its address
                if (keyaddr.equals(from) 
                    && handle.hasSubscriber(subId)) {
                    return;
                }
                // only transfer the suspended buffer, otherwise just
                // drop it and has to recover from end-to-end ack/nack
                if (handle.hasSubscriber(subId)) {
                    if (handle.isSuspended()) {
                        buffer = handle.removeSubscriber(subId);
                    }
                    else {
                        handle.removeSubscriber(subId);
                    }
                }
                if (buffer != null) break;  // must have only one
            }
            // now create the buffer handle
            handle = (BufferHandle) handles.get(from);
            if (handle == null) {
                handle = new GroupBufferHandle(
                    pubId, from, _state._suspendLimit);
                handles.put(from, handle);
            }
            handle.addSubscriber(subId, policy, buffer);
        }
    }

    /** Cancel all the buffers for specified publisher. */
    public void removePublisher(Id pubId)
    {
        
    }

    /** Cancel all the buffers for specified subscriber. */
    public void removeSubscriber(Id subId)
    {
        
    }

    /** Stop delivering the event to subscriber, but continue receiving events. */
    public void suspendSubscriber(Object subkey)
    {
        _outbufman.suspendSubscriber(subkey);
    }

    /** Restart delivering events to specified subscriber. */
    public void resumeSubscriber(Object subkey)
    {
        _outbufman.resumeSubscriber(subkey);
    }

    /** Return the string representation of this protocol. */
    public String toString()
    {
        StringBuffer ret = new StringBuffer();
        synchronized (_state)
        {
            Iterator itr = _state._buffers.values().iterator();
            while (itr.hasNext())
            {
                Hashtable handles = (Hashtable) itr.next();
                Iterator itr2 = handles.values().iterator();
                while (itr2.hasNext())
                {
                    ret.append(itr2.next());
                }
            }
        }
        return ret.toString();
    }

    //=======================================================================
    // Implementation of Protocol interface.
    //=======================================================================

    public void putEvent(Event evt, Address from)
    {
        Id pubId = ((PackEvent) evt).getPubId();
        Id subId = ((PackEvent) evt).getSubId();
        assertPubRegistered(pubId);
        try
        {
            _inbufman.putEvent(pubId, subId, evt);
        }
        catch (BufferFullException bfe)
        {
            _log.error(bfe);
        }
    }

    public Event getEvent()
        throws InterruptedException
    {
        return _outbufman.getEvent();
    }

    public void update(ProtocolEvent evt)
    {
        // default to do nothing
    }
    
    //=======================================================================
    // Private helper methods.
    //=======================================================================

    private void assertPubRegistered(Id pubId)
    {
        Object obj = null;
        synchronized (_state)
        {
            obj = _state._buffers.get(pubId);
        }
        if (obj == null)
        {
            throw new IllegalArgumentException("Buffer " + 
                pubId + " does not exist.");
        }
    }
}
