/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.pubsub.client;


import serp.util.Options;
import org.apache.log4j.Logger;
import java.io.FileInputStream;
import java.util.List;

import solar.api.Id;
import solar.api.Event;
import solar.api.Address;
import solar.service.SolarServices;
import solar.service.ServiceManager;
import solar.service.dispatch.DispatchService;
import solar.service.pubsub.PubSubService;
import solar.service.pubsub.Publisher;
import solar.service.pubsub.Subscriber;
import solar.Planet;


/**
 *  The facade class to interact with Solar system for Pub/Sub service.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: PubSubClient.java,v 1.6 2003/10/12 00:58:39 glchen Exp $
 */
public class PubSubClient 
{
    protected static Logger _log = Logger.getLogger("solar");
    
    private DispatchService _dispatch;
    private PubSubService _pubsub;

    
    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public PubSubClient(String host, int port)
        throws Exception
    {
        Options opts = getDefaultOptions();
        opts.setProperty("service.pubsub.server", 
            host + ":" + Integer.toString(port));
        construct(opts);
    }

    public PubSubClient(String host, int port, Options opts)
        throws Exception
    {
        String debug = opts.getProperty("debug");
        String layout = opts.getProperty("layout");
        String classname = opts.getProperty("service.edge_transport.classname");
        String newport = opts.getProperty("service.edge_transport.port");
        String serialize = opts.getProperty("serialize");
        if (serialize == null)
        {
            serialize = opts.getProperty("service.edge_transport.serialize");
        }
        if (opts.getProperty("cfg") == null)
        {
            opts = getDefaultOptions();
        }
        opts.setProperty("debug", debug);
        opts.setProperty("layout", layout);
        if (classname != null)
            opts.setProperty("service.edge_transport.classname", classname);
        if (newport != null)
            opts.setProperty("service.edge_transport.port", newport);
        opts.setProperty("service.edge_transport.serialize", serialize);
        opts.setProperty("service.pubsub.server", 
            host + ":" + Integer.toString(port));
        construct(opts);
    }

    public PubSubClient(String cfg)
        throws Exception
    {
        Options opts = new Options();
        opts.load(new FileInputStream(cfg));
        construct(opts);
    }

    public PubSubClient(Options opts)
        throws Exception
    {
        construct(opts);
    }

    private Options getDefaultOptions()
    {
        Options opts = new Options();
        opts.setProperty("service.dispatch.classname", 
            "solar.service.dispatch.SolarDispatchService");
        opts.setProperty("service.dispatch.transport", "edge_transport");

        opts.setProperty("service.pubsub.classname", 
            "solar.service.pubsub.ClientPubSubService");
        opts.setProperty("service.pubsub.rpc", "edge_rpc");
        opts.setProperty("service.pubsub.transport", "edge_transport");

        opts.setProperty("service.edge_rpc.classname",
            "solar.service.rpc.SolarRpcService");
        opts.setProperty("service.edge_rpc.transport", "edge_transport");
            
        opts.setProperty("service.edge_transport.classname", 
            "solar.service.transport.UdpTransportService");
        opts.setProperty("service.edge_transport.port", 0);
        return opts;
    }

    /** Common operations needed by constructors. */
    private void construct(Options opts)
        throws Exception
    {
        // set the logging verbose level
        Planet.configLogger(_log, opts);

        // init service manager and relavant services
        ServiceManager sm = SolarServices.getInstance();
        sm.init(opts);
        _dispatch = (DispatchService) sm.getService("dispatch");
        _pubsub = (PubSubService) sm.getService("pubsub");
    }


    //=======================================================================
    // Public interface methods.
    //=======================================================================

    public void addPublisher(Id pubId, Publisher publisher)
        throws Exception
    {
        _pubsub.addPublisher(pubId, publisher);
    }

    public void addSubscriber(Id subId, Subscriber subscriber)
        throws Exception
    {
        _pubsub.addSubscriber(subId, subscriber);
    }
    
    public void subscribe(Id pubId, Id subId, Object extra)
        throws Exception
    {
        _pubsub.subscribe(pubId, subId, extra);
    }

    public void publish(Id pubId, Event evt)
        throws Exception
    {
        _pubsub.publish(pubId, evt);
    }


    
    //=======================================================================
    // Main entrance for testing purpose.
    //=======================================================================

    /**
     *  Start the client and make name lookups. Command line
     *  usage:
     *  <p>
     *  -host tahoe -port 5470
     */
    public static void main(String[] args)
        throws Exception
    {
        Options opts = new Options();
        opts.setFromCmdLine(args);

        String host = opts.getProperty("host", "hogback");
        int port = opts.getIntProperty("port", 5470);
        PubSubClient client = new PubSubClient(host, port);
    }

}
