/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.service.transport;


import org.apache.log4j.Logger;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.EOFException;
import java.net.Socket;
import java.net.ServerSocket;
import java.net.InetAddress;
import java.net.BindException;
import java.net.SocketException;
import java.util.Random;
import java.util.List;
import java.util.Vector;
import java.util.Iterator;
import java.util.Hashtable;

import solar.api.Id;
import solar.api.Adu;
import solar.api.Address;
import solar.api.AduFactory;
import solar.util.NetUtil;


/**
 *  The TCP transport layer, converting ADU to bytes before sending.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: TcpTransport.java,v 1.16 2003/10/12 21:36:18 glchen Exp $
 */
public class TcpTransport 
    implements Transport, Runnable
{
    /** Logging facility. */
    protected static Logger _log;

    /** Particular interface to bind to. */
    protected String _hostname;
    
    /** The listening server port. */
    protected int _port;

    /** The listening server socket. */
    protected ServerSocket _ssock = null;

    /** The server thread accepting new connection. */
    protected Thread _server;

    /** The table of live connections. */
    protected Hashtable _flows;

    /** The list of TCP listeners. */
    protected List _listeners = new Vector();

    /** The local address for TCP. */
    protected Address _laddr;

    /** The factory to parse received byte array into ADU. */
    protected AduFactory _aduFactory;
    

    //=======================================================================
    // Constructors.
    //=======================================================================

    /** Constructs a new TCP transport, which listens on an available port. */
    public TcpTransport()
        throws Exception
    {
        this(0);
    }
    
    /** Constructs a new TCP transport, which listens on a specified port. */
    public TcpTransport(int port)
        throws Exception
    {
        _log = Logger.getLogger(TcpTransport.class);
        _port = port;
        _flows = new Hashtable();
        start();
    }

    public TcpTransport(String hostname, int port)
        throws Exception
    {
        _log = Logger.getLogger(TcpTransport.class);
        _hostname = hostname;
        _port = port;
        _flows = new Hashtable();
        start();
    }

    /**  Start the TCP transport. */
    private void start()
        throws Exception
    {
        _ssock = new ServerSocket();
        _hostname = (_hostname == null) ? "localhost" : _hostname;
        InetAddress tobind = NetUtil.getInetAddress(_hostname);
        Random ran = new Random(System.currentTimeMillis());
        if (_port > 0)
        {
            //_ssock = new ServerSocket(_port);
            _ssock.bind(new java.net.InetSocketAddress(tobind, _port));
        }
        else
        {
            _port = 1050 + ran.nextInt(500);
            while(true) 
            {
                try 
                {
                    //_ssock = new ServerSocket(_port);
                    _ssock.bind(new java.net.InetSocketAddress(tobind, _port));
                    break;
                }
                catch (BindException be)
                {
                    _port += ran.nextInt(100);
                }
            }
        }
        _server = new Thread(this);
//        _server.setDaemon(true);
        _server.start();
        /*
        try
        {
            Socket sk = new Socket("checkip.dyndns.org", 8245);
            _laddr = new SocketAddress(sk.getLocalAddress(), _port);
            sk.close();
        }
        catch (Exception ex)
        {
            _laddr = new SocketAddress("localhost", _port);
        //}*/
        _laddr = new SocketAddress(_hostname, _port);
        _log.info("TCP transport started at local address: " + _laddr);
    }

    //=======================================================================
    // Implementation of solar.service.transport.Transport interface.
    //=======================================================================

    public void setAduFactory(AduFactory factory)
    {
        _aduFactory = factory;
    }

    public Address getLocalAddress()
    {
        return _laddr;
    }

    public Address getAddressByKey(String key)
        throws TransportException
    {
        throw new TransportException("DHT function not supported.");
    }

    public void send(byte[] data, Address dest)
        throws TransportException
    {
        send(data, dest, null);
    }

    public void send(byte[] data, Address dest, Id multiplex)
        throws TransportException
    {
        SocketAddress rdest = (SocketAddress) dest;
        TcpConnection conn = getConnection(rdest);
        try
        {
            if (conn == null)
            {
                conn = openConnection(rdest);
            }
            Header header = new Header(multiplex);
            header.setSource(_laddr);
            header.setDestination(dest);
            conn.send(data, header);
        }
        catch (Exception ex)
        {
            throw new TransportException("Error sending TCP data.", ex);
        }
    }

    public void send(Adu adu, Address dest)
        throws TransportException
    {
        send(adu, dest, null);
    }

    public void send(Adu adu, Address dest, Id multiplex)
        throws TransportException
    {
        send(adu.toBytes(), dest, multiplex);
    }

    public void addTransportListener(TransportListener listener)
    {
        _listeners.add(listener);
    }

    public void removeTransportListener(TransportListener listener)
    {
        _listeners.remove(listener);
    }

    public void stop()
    {
        if (_ssock != null)
        {
            try
            {
                _ssock.close();
                _ssock = null;
                _server.interrupt();
            }
            catch (Exception ex) {}
        }
        clearConnections();
        _laddr = null;
        _listeners.clear();
    }

    //=======================================================================
    // Public interface methods.
    //=======================================================================

    /** Create a TcpConenction based on the socket and peer address. */
    protected TcpConnection makeTcpConnection(Socket socket, SocketAddress peer)
        throws Exception
    {
        return new TcpConnection(false, socket, peer);
    }
    
    /** Return the port number listened by this transport. */
    public int getPort()
    {
        return _port;
    }

    /** Return the logger of this transport. */
    public Logger getLogger()
    {
        return _log;
    }

    /** Return the factory producing Adu from byte array. */
    public AduFactory getAduFactory()
    {
        return _aduFactory;
    }

    /** Notify transport listeners about received data. */
    public void notifyDataReceived(Adu adu, Header header)
    {
        Iterator itr = _listeners.iterator();
        while (itr.hasNext())
        {
            TransportListener tl = (TransportListener) itr.next();
            tl.dataReceived(adu, header);
        }
    }

    /** Notify transport listeners about new connection. */
    public void notifyConnectionOpened(SocketAddress peer)
    {
        Iterator itr = _listeners.iterator();
        while (itr.hasNext())
        {
            TransportListener tl = (TransportListener) itr.next();
            tl.connectionOpened(peer);
        }
    }

    /** Notify transport listeners about closed connection. */
    public void notifyConnectionClosed(SocketAddress peer)
    {
        removeConnection(peer);
        
        Iterator itr = _listeners.iterator();
        while (itr.hasNext())
        {
            TransportListener tl = (TransportListener) itr.next();
            tl.connectionClosed(peer);
        }
    }
    
    //=======================================================================
    // Private helper methods.
    //=======================================================================
    
    /** Get the connection for specified peer address. */
    protected synchronized TcpConnection getConnection(SocketAddress peer)
    {
        return (TcpConnection) _flows.get(peer);
    }

    /** Add a new connection to peer address. */
    protected synchronized void addConnection(
        SocketAddress peer, TcpConnection conn)
    {
        _flows.put(peer, conn);
    }

    /** Remove a connection to a peer address. */
    protected synchronized void removeConnection(SocketAddress peer)
    {
        TcpConnection conn = (TcpConnection) _flows.remove(peer);
        if (conn != null)
            conn.destroy();
    }

    /** Open a new connection to a specified address. */
    protected TcpConnection openConnection(SocketAddress dest)
        throws Exception
    {
        Socket sock = new Socket(dest.getInetAddress(), dest.getPort());
        TcpConnection conn = makeTcpConnection(sock, dest);
		conn.init(this);
		conn.sendLocalAddress( (SocketAddress) _laddr );
		addConnection(dest, conn);
		notifyConnectionOpened(dest);
        return conn;
    }

    /** Remove all the connections. */
    protected synchronized void clearConnections()
    {
        Iterator itr = _flows.values().iterator();
        while (itr.hasNext())
        {
            TcpConnection conn = (TcpConnection) itr.next();
            conn.destroy();
        }
        _flows.clear();
    }


    //=======================================================================
    // Implementation of java.lang.Runnable interface.
    //=======================================================================

    public void run()
    {
        Socket sock = null;
        TcpConnection conn = null;
        SocketAddress peer = null;
        
        while (_ssock != null)
        {
            try
            {
                sock = _ssock.accept();
                conn = makeTcpConnection(sock, null);
                peer = conn.receivePeerAddress();
                if (getConnection(peer) != null)
                {
                    conn.destroy();
                }
                else
                {
                    conn.init(this);
                    addConnection(peer, conn);
                    notifyConnectionOpened(peer);
                }
            }
            catch(SocketException se) 
            {
                _log.warn(se);
                if (conn != null)
                    conn.destroy();
                if (_ssock == null)
                    break;
            }
            catch (Exception ex)
            {
                _log.error(ex);
            }
        }
        _log.error("TCP transport is stopped due to close of server socket.");
    }

}
