/*
 * ============================================================================
 *                   The Solar Software License, Version 1.0
 * ============================================================================
 *
 *       Copyright (C) 2002-2003 Dartmouth College. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of  source code must  retain the above copyright  notice,
 *    this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * 3. The  names  "Solar"  and   "Dartmouth  College"  must  not  be  used   to 
 *    endorse  or   promote   products  derived   from this  software   without
 *    prior  written   permission.   For  written   permission,  please contact 
 *    solar@cs.dartmouth.edu.
 * 
 * 4. Products  derived from this software may not  be called "Solar", nor may
 *    "Solar" appear  in their name,  without prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 * DARTMOUTH COLLEGE OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY  DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR   CONSEQUENTIAL   DAMAGES   (INCLUDING
 * BUT  NOT  LIMITED  TO, PROCUREMENT OF SUBSTITUTE  GOODS  OR  SERVICES;  LOSS
 * OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 * ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 * (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * This software  consists of voluntary contributions made  by many individuals
 * on behalf of the Solar project at Dartmouth College. For more information on
 * the Solar project, please see <http://www.cs.dartmouth.edu/~solar/>.

 */

package solar.test;


import cern.jet.random.Uniform;
import cern.jet.random.Exponential;
import cern.jet.random.engine.MersenneTwister;
import edu.cornell.lassp.houle.RngPack.Ranlux;
import org.apache.log4j.Logger;
import serp.util.Options;
import java.io.ObjectOutputStream;
import java.io.ByteArrayOutputStream;

import solar.Planet;
import solar.api.Adu;
import solar.api.Address;
import solar.impl.StringAdu;
import solar.service.transport.Header;
import solar.service.transport.Transport;
import solar.service.transport.SocketAddress;
import solar.service.transport.TransportListener;
import solar.service.transport.TcpObjectTransport;


/**
 *  This class can be used to produce background network traffic.
 *  
 *  @author     Guanling Chen
 *  @version    $Id: GroundHog.java,v 1.2 2003/10/13 03:13:06 glchen Exp $
 */
public class GroundHog 
    implements TransportListener
{
    protected static Logger _log = Logger.getLogger("solar");
    protected static String[] _hex = {"0", "1", "2", "3", "4", "5", "6", "7",
        "8", "9", "a", "b", "c", "d", "e", "f"};
    
    protected Options _opts;
    protected Uniform _unigen;
    protected Transport _transport;
    
    //=======================================================================
    // Constructors.
    //=======================================================================
    
    public GroundHog(Options opts)
    {
        _opts = opts;
        _unigen = new Uniform(new MersenneTwister());
    }

    //=======================================================================
    // Public interface methods.
    //=======================================================================

    static int counter = 0;    
    public void dataReceived(Adu adu, Header header)
    {

        try {
            _log.debug((++counter) + " evtlen [" + adu.toString().length() 
                + "] adusize [" + getAduSize(adu) + "]");
        } catch (Exception ex) {}
    }

    public void connectionOpened(Address peer) {}

    public void connectionClosed(Address peer) {}

    //=======================================================================
    // Public interface methods.
    //=======================================================================

    public void run()
        throws Exception
    {
        int port = _opts.getIntProperty("port", 5460);
        String hostname = _opts.getProperty("host", "localhost");
        _transport = new TcpObjectTransport(hostname, port);
        _transport.addTransportListener(this);
        boolean server = _opts.getBooleanProperty("server", true);
        if (!server) {
            String serverhost = _opts.getProperty("serverhost", "localhost");
            int serverport = _opts.getIntProperty("serverport", 5460);
            Address dest = new SocketAddress(serverhost, serverport);

            int evtlen = _opts.getIntProperty("evtlen", 50);
            String tra = _opts.getProperty("traffic", "uni");
            int interval = _opts.getIntProperty("interval", 3000);
            if (tra.equals("uni")) {
                goUniform(evtlen, dest, interval);
            }
            else if (tra.equals("exp")) {
                double lambda = _opts.getDoubleProperty("lambda", 0.1);
                goExponential(evtlen, dest, lambda, interval);
            }
            else {
                throw new Exception("Unsupported traffic type: " + tra);
            }
        }
    }

    private void goUniform(int evtlen, Address dest, int interval)
        throws Exception
    {
        int cnt = 0;
        while(true)
        {
            long start = System.currentTimeMillis();
            Adu adu = getAdu(evtlen);
            _transport.send(adu, dest);
            long sleep = interval - (System.currentTimeMillis() - start);
            if (sleep > 0) {
                Thread.sleep(sleep);
            }
        }
    }

    private void goExponential(int evtlen, Address dest, double lambda, int interval)
        throws Exception
    {
        Exponential exp = new Exponential(lambda, new Ranlux());
        int cnt = 0;
        boolean on = true;
        while (true) {
            long fliptime = getNextInterval(exp);
            _log.debug("exp: " + fliptime);
            if (on) {
                long flipstart = System.currentTimeMillis();
                while (System.currentTimeMillis() - flipstart < fliptime) {
                    long start = System.currentTimeMillis();
                    Adu adu = getAdu(evtlen);
                    _transport.send(adu, dest);
                    long sleep = interval - (System.currentTimeMillis() - start);
                    if (sleep > 0) {
                        Thread.sleep(sleep);
                    }
                }
            }
            else {
                Thread.sleep(fliptime);
            }
            on = !on;
        }
    }

    private long getNextInterval(Exponential exp)
    {
        return (long) exp.nextDouble()*1000;
    }
    
    private Adu getAdu(int evtlen)
    {
        StringBuffer ret = new StringBuffer();
        for (int i=0; i<evtlen; i++) {
            ret.append(_hex[_unigen.nextIntFromTo(0, _hex.length-1)]);
        }
        return new StringAdu(ret.toString());
    }

    private int getAduSize(Adu adu)
        throws Exception
    {
        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        ObjectOutputStream out = new ObjectOutputStream(bout);
        out.writeObject(adu);
        int ret = bout.size();
        out.close();
        bout.close();
        return ret;
    }

    //=======================================================================
    // Main entrance.
    //=======================================================================
    
    public static void main(String[] args)
        throws Exception
    {
        Options opts = new Options();
        opts.setFromCmdLine(args);
        Planet.configLogger(_log, opts);
        new GroundHog(opts).run();
    }
}
